import numpy as np
import subprocess, sys, os.path
from itertools import *
from fastlmm.pyplink.snpset import *
from fastlmm.pyplink.altset_list import *
import pandas as pd
import logging
import warnings

WRAPPED_PLINK_PARSER_PRESENT = None


def decide_once_on_plink_reader():
    # This is now done in a method, instead of at the top of the file, so that messages can be re-directed to the appropriate stream.
    # (Usually messages go to stdout, but when the code is run on Hadoop, they are sent to stderr)

    global WRAPPED_PLINK_PARSER_PRESENT
    if WRAPPED_PLINK_PARSER_PRESENT == None:
        # attempt to import wrapped plink parser
        try:
            from bed_reader import read_f32, read_f64

            WRAPPED_PLINK_PARSER_PRESENT = (
                True  #!!!does the standardizer work without c++
            )
            logging.info("using Rust-based plink parser")
        except Exception as detail:
            logging.warn(detail)
            WRAPPED_PLINK_PARSER_PRESENT = False


class Bed(object):
    """
    This is an obsolete class that does random-access reads of a BED file. For examples of its use see its 'read' method.
    """

    def __init__(self, basefilename):
        """
        basefilename    : string of the basename of [basename].bed, [basename].bim,
                          and [basename].fam
        """
        warnings.warn(
            "This old Bed reader is deprecated"
            " Use PySnpTools or bed-reader"
            " instead.",
            DeprecationWarning,
        )

        self._ran_once = False
        self._filepointer = None

        self.basefilename = basefilename

    #!! similar code in fastlmm
    def __repr__(self):
        return "{0}('{1}')".format(self.__class__.__name__, self.basefilename)

    @property
    def snp_to_index(self):
        self.run_once()
        return self._snp_to_index

    def run_once(self):
        if self._ran_once:
            return
        self._ran_once = True

        famfile = self.basefilename + ".fam"
        bimfile = self.basefilename + ".bim"

        logging.info("Loading fam file {0}".format(famfile))
        self._original_iids = np.loadtxt(
            famfile, dtype="str", usecols=(0, 1), comments=None
        )
        logging.info("Loading bim file {0}".format(bimfile))

        self.bimfields = pd.read_csv(
            bimfile,
            delimiter="\s",
            usecols=(0, 1, 2, 3),
            header=None,
            index_col=False,
            engine="python",
        )
        self.rs = np.array(self.bimfields[1].tolist(), dtype="str")
        self.pos = self.bimfields[[0, 2, 3]].values
        self._snp_to_index = {}
        logging.info("indexing snps")
        for i in range(self.snp_count):
            snp = self.rs[i]
            if snp in self._snp_to_index:
                raise Exception(
                    "Expect snp to appear in bim file only once. ({0})".format(snp)
                )
            self._snp_to_index[snp] = i

        bedfile = self.basefilename + ".bed"
        self._filepointer = open(bedfile, "rb")
        mode = self._filepointer.read(2)
        if mode != b"l\x1b":
            raise Exception("No valid binary BED file")
        mode = self._filepointer.read(1)  # \x01 = SNP major \x00 = individual major
        if mode != b"\x01":
            raise Exception("only SNP-major is implemented")
        logging.info("bed file is open {0}".format(bedfile))
        return self

    def __del__(self):
        if (
            self._filepointer is not None
        ):  # we need to test this because Python doesn't guarantee that __init__ was fully run
            self._filepointer.close()

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because only uses original inputs
        copier.input(self.basefilename + ".bed")
        copier.input(self.basefilename + ".bim")
        copier.input(self.basefilename + ".fam")

    @property
    def snp_count(self):
        self.run_once()
        return len(self.bimfields)

    def read(
        self, snp_set=AllSnps(), order="F", dtype=np.float64, force_python_only=False
    ):
        """
        Input: a snp_set. Choices include
            AllSnps() [the default],
            PositionRange(snpIndex,nSNPs)
            SnpAndSetName(groupname,snplist),

        Output dictionary:
        'rs'     : [S] array rs-numbers
        'pos'    : [S*3] array of positions [chromosome, genetic dist, basepair dist]
        'snps'   : [N*S] array of snp-data
        'iid'    : [N*2] array of family IDs and individual IDs

        Examples:

        >>> from fastlmm.util import example_file # Download and return local file name
        >>> bed_file = example_file('tests/datasets/all_chr.maf0.001.N300.*','*.bed').replace('.bed','')
        >>> bed = Bed(bed_file)
        >>> ret = bed.read()
        >>> len(ret['rs'])
        1015
        >>> ret = bed.read(AllSnps())
        >>> len(ret['rs'])
        1015
        >>> ret = bed.read(SnpAndSetName('someset',['23_9','23_2']))
        >>> ",".join(ret['rs'])
        '23_9,23_2'
        >>> ret = bed.read(PositionRange(0,10))
        >>> ",".join(ret['rs'])
        '1_12,1_34,1_10,1_35,1_28,1_25,1_36,1_39,1_4,1_13'


        >>> altset_list1 = SnpAndSetNameCollection(example_file('tests/datasets/set_input.small.txt')) # get the list of snpsets defined in the file
        >>> altset_list2 = Subset(altset_list1,['set1','set5'])                       # only use a subset of those snpsets
        >>> altset_list3 = MinMaxSetSize(altset_list2, minsetsize=2, maxsetsize=15)   # only use the subset of subsets that contain between 2 & 15 snps (inclusive)
        >>> bed_file = example_file('tests/datasets/all_chr.maf0.001.N300.*','*.bed').replace('.bed','')
        >>> bed = Bed(bed_file)
        >>> altsetlist_plusbed = altset_list3.addbed(bed)                             # apply altset_list3 to this bed file
        >>> len(altsetlist_plusbed)                                                   # tell how many snpsets there will be
        1
        >>> snpset_plusbed = list(altsetlist_plusbed)[0]
        >>> str(snpset_plusbed)                                                       # the name of the snpset
        'set5'
        >>> len(snpset_plusbed)                                                       # the number of snps in the snpset
        13
        >>> ret = snpset_plusbed.read()
        >>> ",".join(ret['rs'])
        '5_12,5_28,5_32,5_5,5_11,5_1,5_9,5_3,5_19,5_7,5_21,5_15,5_23'

        """
        self.run_once()
        snpset_withbed = snp_set.addbed(self)
        return self.read_with_specification(
            snpset_withbed,
            order=order,
            dtype=dtype,
            force_python_only=force_python_only,
        )

    ##!! This property is ugly
    @property
    def ind_used(self):
        # doesn't need to self.run_once() because only uses original inputs
        return self._ind_used

    @ind_used.setter
    def ind_used(self, value):
        """
        Tell the Bed reader to return data for only a subset (perhaps proper) of the individuals in a particular order
        e.g. 2,10,0 says to return data for three users: the user at index position 2, the user at index position 10, and the user at index position 0.
        """
        # doesn't need to self.run_once() because only uses original inputs
        self._ind_used = value

    @property
    def original_iids(self):
        self.run_once()
        return self._original_iids

    def counts_and_indexes(self, snpset_withbbed):
        iid_count_in = len(self.original_iids)
        snp_count_in = self.snp_count
        if hasattr(self, "_ind_used"):
            iid_count_out = len(self.ind_used)
            iid_index_out = self.ind_used
        else:
            iid_count_out = iid_count_in
            iid_index_out = list(range(0, iid_count_in))
        snp_count_out = len(snpset_withbbed)
        snp_index_out = list(
            snpset_withbbed
        )  # make a copy, in case it's in some strange format, such as HDF5
        return (
            iid_count_in,
            iid_count_out,
            iid_index_out,
            snp_count_in,
            snp_count_out,
            snp_index_out,
        )

    @staticmethod
    def read_with_specification(
        snpset_withbbed, order="F", dtype=np.float64, force_python_only=False
    ):
        # doesn't need to self.run_once() because it is static
        decide_once_on_plink_reader()
        global WRAPPED_PLINK_PARSER_PRESENT

        bed = snpset_withbbed.bed
        (
            iid_count_in,
            iid_count_out,
            iid_index_out,
            snp_count_in,
            snp_count_out,
            snp_index_out,
        ) = bed.counts_and_indexes(snpset_withbbed)

        if WRAPPED_PLINK_PARSER_PRESENT and not force_python_only:
            from bed_reader import open_bed

            bed_fn = bed.basefilename + ".bed"
            count_A1 = False

            if dtype == np.float64:
                dtype = "float64"
            elif dtype == np.float32:
                dtype = "float32"
            else:
                raise ValueError(
                    f"dtype '{val.dtype}' not known, only "
                    + "'int8', 'float32', and 'float64' are allowed."
                )

            with open_bed(
                bed_fn,
                iid_count=iid_count_in,
                sid_count=snp_count_in,
                count_A1=count_A1,
            ) as bed2:
                SNPs = bed2.read(np.s_[iid_index_out, snp_index_out], dtype=dtype)

        else:
            # An earlier version of this code had a way to read consecutive SNPs of code in one read. May want
            # to add that ability back to the code.
            # Also, note that reading with python will often result in non-contiguous memory, so the python standardizers will automatically be used, too.
            logging.warn("using pure python plink parser (might be much slower!!)")
            SNPs = np.zeros(
                ((int(np.ceil(0.25 * iid_count_in)) * 4), snp_count_out),
                order=order,
                dtype=dtype,
            )  # allocate it a little big
            for SNPsIndex, bimIndex in enumerate(snpset_withbbed):
                startbit = int(np.ceil(0.25 * iid_count_in) * bimIndex + 3)
                bed._filepointer.seek(startbit)
                nbyte = int(np.ceil(0.25 * iid_count_in))
                bytes = np.array(bytearray(bed._filepointer.read(nbyte))).reshape(
                    (int(np.ceil(0.25 * iid_count_in)), 1), order="F"
                )

                SNPs[3::4, SNPsIndex : SNPsIndex + 1][bytes >= 64] = np.nan
                SNPs[3::4, SNPsIndex : SNPsIndex + 1][bytes >= 128] = 1
                SNPs[3::4, SNPsIndex : SNPsIndex + 1][bytes >= 192] = 2
                bytes = np.mod(bytes, 64)
                SNPs[2::4, SNPsIndex : SNPsIndex + 1][bytes >= 16] = np.nan
                SNPs[2::4, SNPsIndex : SNPsIndex + 1][bytes >= 32] = 1
                SNPs[2::4, SNPsIndex : SNPsIndex + 1][bytes >= 48] = 2
                bytes = np.mod(bytes, 16)
                SNPs[1::4, SNPsIndex : SNPsIndex + 1][bytes >= 4] = np.nan
                SNPs[1::4, SNPsIndex : SNPsIndex + 1][bytes >= 8] = 1
                SNPs[1::4, SNPsIndex : SNPsIndex + 1][bytes >= 12] = 2
                bytes = np.mod(bytes, 4)
                SNPs[0::4, SNPsIndex : SNPsIndex + 1][bytes >= 1] = np.nan
                SNPs[0::4, SNPsIndex : SNPsIndex + 1][bytes >= 2] = 1
                SNPs[0::4, SNPsIndex : SNPsIndex + 1][bytes >= 3] = 2
            SNPs = SNPs[iid_index_out, :]  # reorder or trim any extra allocation

        ret = {
            "rs": bed.rs[snp_index_out],
            "pos": bed.pos[snp_index_out, :],
            "snps": SNPs,
            "iid": bed.original_iids[iid_index_out, :],
        }
        return ret


if __name__ == "__main__":
    if False:
        from fastlmm.util import example_file  # Download and return local file name

        bed_file = example_file("tests/datasets/all_chr.maf0.001.N300.*", "*.bed")
        bed = Bed(bed_file)
        ret = bed.read()
        len(ret["rs"])
        # 1015
        ret = bed.read(AllSnps())
        len(ret["rs"])
        # 1015
        ret = bed.read(SnpAndSetName("someset", ["23_9", "23_2"]))
        ",".join(ret["rs"])
        #'23_9,23_2'
        ret = bed.read(PositionRange(0, 10))
        ",".join(ret["rs"])
        #'1_12,1_34,1_10,1_35,1_28,1_25,1_36,1_39,1_4,1_13'

    logging.basicConfig(level=logging.WARNING)
    import doctest

    doctest.testmod()



import numpy as np
import logging
import unittest
import os.path
import time
import sys
import doctest
import pandas as pd

from fastlmm.association import snp_set
import fastlmm.pyplink.plink as plink
from fastlmm.feature_selection.test import TestFeatureSelection
from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc, LocalInParts
import fastlmm.util.util as ut

tolerance = 1e-4


class TestSnpSet(unittest.TestCase):

    #!!created a Expect Durbin, too

    @classmethod
    def setUpClass(self):

        from pysnptools.util import create_directory_if_necessary

        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.currentFolder = os.path.dirname(os.path.realpath(__file__))

    tempout_dir = "tempout/snp_set"

    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name)
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    def test_one(self):
        logging.info("TestSnpSet test_one")

        fn = "lrt_one_kernel_fixed_mixed_effect_linear_qqfit.N300.txt"
        tmpOutfile = self.file_name(fn)
        referenceOutfile = self._referenceOutfile(fn)

        result_dataframe = snp_set(
            test_snps=self.currentFolder
            + "/../../../tests/datasets/all_chr.maf0.001.N300",
            set_list=self.currentFolder + "/../../../tests/datasets/set_input.23.txt",
            pheno=self.currentFolder
            + "/../../../tests/datasets/phenSynthFrom22.23.N300.txt",
            output_file_name=tmpOutfile,
        )

        out, msg = ut.compare_files(tmpOutfile, referenceOutfile, tolerance)
        self.assertTrue(
            out,
            "msg='{0}', ref='{1}', tmp='{2}'".format(msg, referenceOutfile, tmpOutfile),
        )

    def test_two(self):
        logging.info("TestSnpSet test_two")

        fn = "lrt_up_two_kernel_mixed_effect_linear_qqfit.N300.fullrank.txt"
        tmpOutfile = self.file_name(fn)
        referenceOutfile = self._referenceOutfile(fn)

        result_dataframe = snp_set(
            test_snps=self.currentFolder
            + "/../../../tests/datasets/all_chr.maf0.001.N300",
            set_list=self.currentFolder + "/../../../tests/datasets/set_input.23.txt",
            pheno=self.currentFolder
            + "/../../../tests/datasets/phenSynthFrom22.23.N300.txt",
            G0=self.currentFolder
            + "/../../../tests/datasets/all_chr.maf0.001.chr22.23.N300.bed",
            output_file_name=tmpOutfile,
            test="lrt",
        )

        out, msg = ut.compare_files(tmpOutfile, referenceOutfile, tolerance)
        self.assertTrue(
            out, msg
        )  # msg='Files %s and %s are different.' % (tmpOutfile, referenceOutfile))

    def test_three(self):
        logging.info("TestSnpSet test_three")

        fn = "sc_davies_one_kernel_linear_qqfit.N300.txt"
        tmpOutfile = self.file_name(fn)
        referenceOutfile = self._referenceOutfile(fn)

        result_dataframe = snp_set(
            test_snps=self.currentFolder
            + "/../../../tests/datasets/all_chr.maf0.001.N300",
            set_list=self.currentFolder
            + "/../../../tests/datasets/set_input.small.txt",
            pheno=self.currentFolder
            + "/../../../tests/datasets/phenSynthFrom22.23.N300.txt",
            test="sc_davies",
            output_file_name=tmpOutfile,
        )

        out, msg = ut.compare_files(tmpOutfile, referenceOutfile, tolerance)
        self.assertTrue(
            out,
            "msg='{0}', ref='{1}', tmp='{2}'".format(msg, referenceOutfile, tmpOutfile),
        )

    def test_four(self):
        logging.info("TestSnpSet test_four")

        fn = "sc_davies_two_kernel_linear_qqfit.N300.noautoselect.txt"
        tmpOutfile = self.file_name(fn)
        referenceOutfile = self._referenceOutfile(fn)

        result_dataframe = snp_set(
            test_snps=self.currentFolder
            + "/../../../tests/datasets/all_chr.maf0.001.N300",
            set_list=self.currentFolder
            + "/../../../tests/datasets/set_input.small.txt",
            pheno=self.currentFolder
            + "/../../../tests/datasets/phenSynthFrom22.23.N300.randcidorder.txt",
            G0=self.currentFolder
            + "/../../../tests/datasets/all_chr.maf0.001.chr22.23.N300.bed",
            test="sc_davies",
            output_file_name=tmpOutfile,
        )

        out, msg = ut.compare_files(tmpOutfile, referenceOutfile, tolerance)
        self.assertTrue(
            out,
            "msg='{0}', ref='{1}', tmp='{2}'".format(msg, referenceOutfile, tmpOutfile),
        )

    def test_notebook(self):
        # define file names
        test_snps_fn = self.currentFolder + "/../../../tests/datasets/synth/chr1"
        pheno_fn = (
            self.currentFolder + "/../../../tests/datasets/synth/pheno_10_causals.txt"
        )
        cov_fn = self.currentFolder + "/../../../tests/datasets/synth/cov.txt"
        set_list_fn = (
            self.currentFolder + "/../../../tests/datasets/synth/chr1.sets.txt"
        )
        G0_fn = None

        # run SNP-set analysis
        results_df = snp_set(
            test_snps=test_snps_fn,
            G0=G0_fn,
            set_list=set_list_fn,
            pheno=pheno_fn,
            covar=cov_fn,
            test="lrt",
        )
        top = results_df.iloc[0]
        logging.debug(top)
        assert top["SetId"] == "set_65"
        assert np.abs(top["P-value"] - 0.002564) < 0.0001

    def test_doctest(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
        result = doctest.testmod(sys.modules["fastlmm.association.snp_set"])
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def _referenceOutfile(self, _infile):
        import platform

        os_string = platform.platform()
        outfile = os.path.splitext(_infile)[0]

        windows_fn = (
            self.currentFolder + "/../../../tests/expected-Windows/" + outfile + ".txt"
        )
        assert os.path.exists(windows_fn), "Can't find file '{0}'".format(windows_fn)
        debian_fn = (
            self.currentFolder + "/../../../tests/expected-debian/" + outfile + ".txt"
        )
        if not os.path.exists(
            debian_fn
        ):  # If reference file is not in debian folder, look in windows folder
            debian_fn = windows_fn

        if "debian" in os_string or "Linux" in os_string:
            if "Linux" in os_string:
                logging.warning(
                    "comparing to Debian output even though found: %s" % os_string
                )
            return debian_fn
        else:
            if "Windows" not in os_string:
                logging.warning(
                    "comparing to Windows output even though found: %s" % os_string
                )
            return windows_fn


def getTestSuite():

    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestSnpSet)
    return unittest.TestSuite([suite1])


if __name__ == "__main__":
    from fastlmm.association.tests.test_snp_set import TestSnpSet

    suites = unittest.TestSuite([getTestSuite()])

    if True:  # Standard test run
        r = unittest.TextTestRunner(failfast=False)
        ret = r.run(suites)
        assert ret.wasSuccessful()
    else:  # Cluster test run
        from pysnptools.util.mapreduce1.distributabletest import DistributableTest

        runner = HPC(
            10,
            "RR1-N13-09-H44",
            r"\\msr-arrays\Scratch\msr-pool\Scratch_Storage4\Redmond",
            remote_python_parent=r"\\msr-arrays\Scratch\msr-pool\Scratch_Storage4\REDMOND\carlk\Source\carlk\july_7_14\tests\runs\2014-07-24_15_02_02_554725991686\pythonpath",
            update_remote_python_parent=True,
            priority="AboveNormal",
            mkl_num_threads=1,
        )
        runner = Local()
        # runner = LocalMultiProc(taskcount=20,mkl_num_threads=5)
        # runner = LocalInParts(1,2,mkl_num_threads=1) # For debugging the cluster runs
        # runner = Hadoop(100, mapmemory=8*1024, reducememory=8*1024, mkl_num_threads=1, queue="default")
        distributable_test = DistributableTest(suites, "temp_test")
        print(runner.run(distributable_test))

    logging.info("done with testing")



import os
import numpy as np
from .snpset import *
import logging


def readPED(basefilename, delimiter=" ", missing="0", standardize=True, pheno=None):
    """
    read [basefilename].ped and [basefilename].map
    optionally standardize the SNPs and mean impute missing SNPs
    --------------------------------------------------------------------------
    Input:
    basefilename    : string of the basename of [basename].ped and [basename].map
    delimiter       : string (default ' ' space)
    missing         : string indicating a missing genotype (default '0')
    standardize     : boolean
                        True    : mean impute, zero-mean and unit variance the data
                        False   : output the data in 0,1,2 with NaN values
    --------------------------------------------------------------------------
    Output dictionary:
    'rs'     : [S] array rs-numbers,
    'pos'    : [S*3] array of positions [chromosome, genetic dist, basepair dist],
    'snps'   : [N*S] array of snps-data,
    'iid'    : [N*2] array of family IDs and individual IDs
    --------------------------------------------------------------------------
    """
    pedfile = basefilename + ".ped"
    mapfile = basefilename + ".map"
    map = np.loadtxt(mapfile, dtype="str", comments=None)

    rs = map[:, 1]
    pos = np.array(map[:, (0, 2, 3)], dtype="float")
    map = None

    ped = np.loadtxt(pedfile, dtype="str", comments=None)
    iid = ped[:, 0:2]
    snpsstr = ped[:, 6::]
    inan = snpsstr == missing
    snps = np.zeros((snpsstr.shape[0], snpsstr.shape[1] // 2))
    if standardize:
        for i in range(snpsstr.shape[1] // 2):
            snps[inan[:, 2 * i], i] = 0
            vals = snpsstr[~inan[:, 2 * i], 2 * i : 2 * (i + 1)]
            snps[~inan[:, 2 * i], i] += (vals == vals[0, 0]).sum(1)
            snps[~inan[:, 2 * i], i] -= snps[~inan[:, 2 * i], i].mean()
            snps[~inan[:, 2 * i], i] /= snps[~inan[:, 2 * i], i].std()
    else:
        for i in range(snpsstr.shape[1] / 2):
            snps[inan[:, 2 * i], i] = np.nan
            vals = snpsstr[~inan[:, 2 * i], 2 * i : 2 * (i + 1)]
            snps[~inan[:, 2 * i], i] += (vals == vals[0, 0]).sum(1)
    if pheno is not None:
        # TODO: sort and filter SNPs according to pheno
        pass
    ret = {"rs": rs, "pos": pos, "snps": snps, "iid": iid}
    return ret


def readRAW(basefilename, delimiter=" ", missing="0", standardize=True, pheno=None):
    """
    read [basefilename].raw
    optionally standardize the SNPs and mean impute missing SNPs
    --------------------------------------------------------------------------
    Input:
    basefilename    : string of the basename of [basename].ped and [basename].map
    delimiter       : string (default ' ' space)
    missing         : string indicating a missing genotype (default '0')
    standardize     : boolean
                        True    : mean impute, zero-mean and unit variance the data
                        False   : output the data in 0,1,2 with NaN values
    --------------------------------------------------------------------------
    Output dictionary:
    'rs'     : [S] array rs-numbers,
    'pos'    : [S*3] array of positions [chromosome, genetic dist, basepair dist],
    'snps'   : [N*S] array of snps-data,
    'iid'    : [N*2] array of family IDs and individual IDs
    --------------------------------------------------------------------------
    """
    rawfile = basefilename + ".raw"
    # mapfile = basefilename+".map"
    # map = np.loadtxt(mapfile,dtype = 'str',comments=None)

    # rs = map[:,1]
    # pos = np.array(map[:,(0,2,3)],dtype = 'float')
    # map = None
    raw = np.loadtxt(rawfile, dtype="str", comments=None)
    iid = raw[:, 0:2]
    snpsstr = raw[:, 6::]
    inan = snpsstr == missing
    snps = np.zeros((snpsstr.shape[0], snpsstr.shape[1] / 2))
    if standardize:
        for i in range(snpsstr.shape[1] / 2):
            raw[inan[:, 2 * i], i] = 0
            vals = snpsstr[~inan[:, 2 * i], 2 * i : 2 * (i + 1)]
            snps[~inan[:, 2 * i], i] += (vals == vals[0, 0]).sum(1)
            snps[~inan[:, 2 * i], i] -= snps[~inan[:, 2 * i], i].mean()
            snps[~inan[:, 2 * i], i] /= snps[~inan[:, 2 * i], i].std()
    else:
        for i in range(snpsstr.shape[1] / 2):
            snps[inan[:, 2 * i], i] = np.nan
            vals = snpsstr[~inan[:, 2 * i], 2 * i : 2 * (i + 1)]
            snps[~inan[:, 2 * i], i] += (vals == vals[0, 0]).sum(1)
    if pheno is not None:
        # TODO: sort and filter SNPs according to pheno
        pass
    ret = {"rs": rs, "pos": pos, "snps": snps, "iid": iid}
    return ret


def writePhen(phen, filename, missing="9", sep="\t"):
    """
    must contain phen['iid'] and phen['vals']
    """
    N = phen["iid"].shape[0]
    M = phen["vals"].shape[1]
    assert (
        N == phen["vals"].shape[0]
    ), "number of individuals do not match up in phen['vals'] and phen['ids']"
    with open(filename, "w") as f:
        for i in range(0, N):
            tmpstr = phen["iid"][i][0] + sep + phen["iid"][i][1]
            for m in range(0, M):
                tmpstr += sep + str(phen["vals"][i][m])
            tmpstr += "\n"
            f.write(tmpstr)


def readBED(basefilename, snp_set=AllSnps(), order="F"):
    """
    This is a one-shot reader for BED files that internally uses the Bed class. If you need repeated random access to a BED file,
    it is much faster to use the Bed class directly. Such use avoids the need to re-read the associated BIM file.
    """
    from . import snpreader as sr

    bed = sr.Bed(basefilename)
    return bed.read(snp_set, order=order)


def nSnpFromBim(basefilename):
    bim = basefilename + ".bim"
    bimx = np.loadtxt(bim, dtype="str", usecols=(0, 1, 2, 3), comments=None)
    S = bimx.shape[0]
    return S


def findIndex(idsSep, bedidsSep):
    sids = set()
    for i in range(idsSep.shape[0]):
        sids.add(idsSep[i, 0] + "_" + idsSep[i, 1])

    beids = dict()
    for i in range(bedidsSep.shape[0]):
        beids[bedidsSep[i, 0] + "_" + bedidsSep[i, 1]] = i

    inter = sids.intersection(set(beids.keys()))

    return [beids[x] for x in inter]


def filter(phe, bed):
    # for pheno
    index = findIndex(bed["iid"], phe["iid"])
    iid = phe["iid"]
    iid = iid[index, :]
    vals = phe["vals"]
    vals = vals[index]
    phe["iid"] = iid
    phe["vals"] = vals
    # for snp
    index = findIndex(phe["iid"], bed["iid"])
    iid = bed["iid"]
    iid = iid[index, :]
    snps = bed["snps"]
    snps = snps[index, :]
    bed["iid"] = iid
    bed["snps"] = snps
    return


# if __name__ == "__main__":
#    #datadir = "C:\\Users\\lippert\\Projects\\ARIC" ; basefilename = os.path.join(datadir, 'whiteMale')
#    datadir = "testdata" ; basefilename = os.path.join(datadir, 'test')
#    #bed = readBED(basefilename,blocksize = 1,nSNPs = 20000,start = 650000)
#    #bed = readBED(basefilename,blocksize = 1,nSNPs = 20000,start = 0)
#    #phe = loadPhen(basefilename + ".tab")
#    filter(phe, bed)
#    bed = readBED('Gaw14/all', startpos = [1,17.5599], endpos = [1,24.6047])



import fastlmm.association as association
import numpy as np
from sklearn import linear_model
from sklearn import model_selection
from sklearn.model_selection import GridSearchCV

#!! from fastlmm.external.sklearn.metrics.scorer import SCORERS, Scorer
from fastlmm import inference
import fastlmm.util.util as util


class testCV(association.varcomp_test):
    __slots__ = [
        "G0",
        "greater_is_better",
        "model",
        "scoring",
        "n_folds",
        "n_folds_params",
        "verbose",
        "n_jobs_grid",
        "data_permutation",
        "scores_null",
        "params_null",
        "nullModel",
        "altModel",
    ]

    def __init__(
        self,
        Y,
        X=None,
        G0=None,
        appendbias=False,
        model=None,
        n_folds_params=10,
        n_folds=10,
        scoring=None,
        verbose=False,
        n_jobs_grid=1,
        data_permutation=None,
        nested=True,
        greater_is_better=None,
        nullModel=None,
        altModel=None,
    ):  # Note that this code will break as we don't know what it does
        association.varcomp_test.__init__(self, Y=Y, X=X, appendbias=appendbias)

        assert model is None, "Shouldn't we remove this parameter?"
        self.n_jobs_grid = n_jobs_grid
        self.verbose = verbose
        self.n_folds = n_folds
        self.n_folds_params = n_folds_params
        self.G0 = G0
        self.nullModel = nullModel
        self.altModel = altModel

        if data_permutation is None:
            data_permutation = util.generate_permutation(
                self.Y.shape[0], 93828231
            )  # permute with an arbitrary seed
        self.data_permutation = data_permutation

        if "param_grid" in nullModel:
            param_grid = nullModel["param_grid"]
        else:
            param_grid = self._getParamGrid(G0, None, nullModel)

        if scoring is None:
            (self.scoring, self.greater_is_better) = self._getScoring()
        else:
            self.scoring = scoring
            self.greater_is_better = greater_is_better

        model = self._getModel(nullModel, param_grid)
        nested = self._isNested(nullModel)

        self.scores_null, self.params_null = self.score_nestedCV(
            None, model, param_grid, self.nullModel["effect"], nested
        )

    def _getScoring(self):
        if self.nullModel["link"] == "linear":
            scoring = "mse"
            greater_is_better = False
        elif self.nullModel["link"] == "logistic":
            scoring = "binomial"
            greater_is_better = True
        else:
            assert False, "Unknown link function."

        return (scoring, greater_is_better)

    def _getModel(self, modelDesc, param_grid):
        if modelDesc["effect"] == "fixed":
            return self._getFixedEffectModel(
                modelDesc["penalty"], modelDesc["link"], param_grid
            )
        elif modelDesc["effect"] == "mixed":
            return self._getMixedEffectModel(
                modelDesc["link"], modelDesc["approx"], param_grid
            )
        else:
            assert False

    def _getFixedEffectModel(self, penalty, link, param_grid):
        assert penalty in set(["l1", "l2"])

        if link == "linear":
            if penalty == "l2":
                model = linear_model.Ridge(
                    alpha=param_grid["alpha"][0], fit_intercept=True, solver="auto"
                )
            elif penalty == "l1":
                model = linear_model.LassoCV(
                    alphas=param_grid["alpha"],
                    cv=self.n_folds_params,
                    precompute="auto",
                    fit_intercept=True,
                )

        elif link == "logistic":
            model = linear_model.LogisticRegression(
                penalty=penalty,
                dual=False,
                tol=0.0001,
                C=param_grid["C"][0],
                fit_intercept=True,
                intercept_scaling=1,
                class_weight=None,
                random_state=None,
            )
        else:
            assert False, "Unknown link function."

        return model

    def _getMixedEffectModel(self, link, approx, param_grid):
        C = inference.makeBin2KernelAsEstimator(link, approx)
        return C()

    def _isNested(self, modelDesc):
        if (
            modelDesc["effect"] == "fixed"
            and modelDesc["link"] == "linear"
            and modelDesc["penalty"] == "l1"
        ):
            return False
        return True

    def _getParamGrid(self, G0, G1, modelDesc):
        if modelDesc["effect"] == "fixed":
            return self._getParamGridFixedEffectModel(G0, G1, modelDesc["link"])
        elif modelDesc["effect"] == "mixed":
            return self._getParamGridMixedEffectModel(G0, G1)

        assert False

    def _getParamGridFixedEffectModel(self, G0, G1, link):
        if link == "linear":
            param_grid = dict(alpha=0.5 * np.logspace(-5, 5, 20))
        elif link == "logistic":
            param_grid = dict(C=np.logspace(-5, 5, 20))
        else:
            assert False

        return param_grid

    def _getParamGridMixedEffectModel(self, G0, G1):
        param_grid = dict(
            sig02=np.arange(0.0, 2.1, 0.4),
            sig12=np.arange(0.0, 2.1, 0.4),
            sign2=[None],
            beta=[None],
        )

        if G0 is None:
            param_grid["sig02"] = [0.0]
        if G1 is None:
            param_grid["sig12"] = [0.0]

        return param_grid

    def testG(self, G1, type="", i_exclude=None):
        pv = 1.0
        stat = 1.0

        if "param_grid" in self.altModel:
            param_grid = self.altModel["param_grid"]
        else:
            param_grid = self._getParamGrid(self.G0, G1, self.altModel)

        model = self._getModel(self.altModel, param_grid)
        nested = self._isNested(self.altModel)

        scores, params = self.score_nestedCV(
            G1, model, param_grid, self.altModel["effect"], nested
        )

        if self.greater_is_better:
            stat = 2.0 * (scores - self.scores_null).mean()
        else:
            stat = 2.0 * (self.scores_null - scores).mean()

        test = {
            "pv": pv,
            "stat": stat,
            "scores": scores,
            "scores0": self.scores_null,
            "params": params,
            "params0": self.params_null,
            "type": type,  # is it OK to have an object here instead of a name?
        }
        return test

    # the effect parameter should not be used here, but I dont have a better an idea for now
    def score_nestedCV(self, G1, model, param_grid, effect, nested):
        k_fold = model_selection.KFold(n_splits=self.n_folds).split(
            list(range(self.Y.shape[0]))
        )
        i_fold = 0
        scores = np.zeros(self.n_folds)
        params = list()

        for train, test in k_fold:
            (trainData, trainY) = self._packData(G1, train, effect)
            (testData, testY) = self._packData(G1, test, effect)

            if nested:
                clf = GridSearchCV(
                    estimator=model,
                    param_grid=param_grid,
                    n_jobs=self.n_jobs_grid,
                    cv=self.n_folds_params,
                    scoring=self.scoring,
                    verbose=self.verbose,
                )

                clf.fit(trainData, trainY.flatten())

                params.append(clf.best_params_)

                scores[i_fold] = clf.score(
                    testData, testY.flatten(), method_scorer=False
                )
            else:
                model.fit(trainData, trainY.flatten())
                scores[i_fold] = SCORERS[self.scoring](model, testData, testY.flatten())
            i_fold += 1

        return scores, params

    def _packData(self, G1, indices2select, effect):
        if effect == "fixed":
            if G1 is None and self.G0 is None:
                data = self.X[self.data_permutation][indices2select]
            elif G1 is None:
                data = np.column_stack(
                    (
                        self.G0[self.data_permutation][indices2select],
                        self.X[self.data_permutation][indices2select],
                    )
                )
            elif self.G0 is None:
                data = np.column_stack(
                    (
                        G1[self.data_permutation][indices2select],
                        self.X[self.data_permutation][indices2select],
                    )
                )
            else:
                data = np.column_stack(
                    (
                        self.G0[self.data_permutation][indices2select],
                        G1[self.data_permutation][indices2select],
                        self.X[self.data_permutation][indices2select],
                    )
                )
        elif effect == "mixed":
            X = self.X[self.data_permutation]
            if self.G0 is not None:
                G0 = self.G0[self.data_permutation]
            if G1 is not None:
                G1 = G1[self.data_permutation]

            data = []
            for i in range(len(indices2select)):
                lis = [X[indices2select[i]]]
                if G0 is not None:
                    lis.append(G0[indices2select[i]])
                if G1 is not None:
                    lis.append(G1[indices2select[i]])

                data.append(lis)
            else:
                assert False, "Unkown effect type."

        return (data, self.Y[self.data_permutation][indices2select])



from fastlmm import Pr
import scipy.optimize as opt
import numpy as np
from numpy import dot
import scipy.integrate
from scipy.linalg import cholesky, solve_triangular
from fastlmm.external.util.math import (
    check_definite_positiveness,
    check_symmetry,
    mvnormpdf,
    ddot,
    trace2,
    dotd,
)
from fastlmm.external.util.math import stl, stu
from fastlmm.inference.glmm import GLMM_N1K3, GLMM_N3K1
from fastlmm.inference.likelihood import LogitLikelihood, ProbitLikelihood
from fastlmm import Pr
import sys

"""
    Important! Always run test.py in the current folder for unit testing after
    changes have been made.
"""


class DebugUACall(object):
    def __init__(self, innerIters, lastGrad, success, beta, sig02, sig12, sign2):
        self.innerIters = innerIters
        self.lastGrad = lastGrad
        self.success = success
        self.beta = beta
        self.sig02 = sig02
        self.sig12 = sig12
        self.sign2 = sign2


class LaplaceGLMM(object):
    def __init__(self, link):
        self._lasta = None
        self._debugUACalls = []
        self._link = link
        if link == "logistic":
            self._likelihood = LogitLikelihood()
        elif link == "erf":
            self._likelihood = ProbitLikelihood()
        else:
            assert False, "Unknown link function."

    def _calculateW(self, f):
        W = -self._likelihood.hessian_log(f)
        return W

    def _lineSearch(self, a, aprev, m):
        da = a - aprev

        def fobj(alpha):
            a = aprev + alpha * da
            f = self._rdotK(a) + m
            return -(self._likelihood.log(f, self._y) - (f - m).dot(a) / 2.0)

        (alpha, obj, iter, funcalls) = opt.fminbound(
            fobj, 0.0, 1.0, full_output=True, xtol=1e-4, maxfun=25
        )
        obj = -obj
        a = aprev + alpha * da
        f = self._rdotK(a) + m
        return (f, a, obj)

    def _calculateUAGrad(self, f, a):
        grad = self._likelihood.gradient_log(f, self._y) - a
        return grad

    def printDebug(self):
        assert self._debug is True
        from tabulate import tabulate

        iters = [
            self._debugUACalls[i].innerIters for i in range(len(self._debugUACalls))
        ]
        sig02 = [self._debugUACalls[i].sig02 for i in range(len(self._debugUACalls))]
        sig12 = [self._debugUACalls[i].sig12 for i in range(len(self._debugUACalls))]
        sign2 = [self._debugUACalls[i].sign2 for i in range(len(self._debugUACalls))]
        gradMeans = [
            np.mean(abs(self._debugUACalls[i].lastGrad))
            for i in range(len(self._debugUACalls))
        ]

        Pr.prin("*** Update approximation ***")
        Pr.prin("calls: %d" % (len(self._debugUACalls),))

        table = [
            ["", "min", "max", "mean"],
            ["iters", min(iters), max(iters), np.mean(iters)],
            ["|grad|_{mean}", min(gradMeans), max(gradMeans), np.mean(gradMeans)],
            ["sig01", min(sig02), max(sig02), np.mean(sig02)],
            ["sig11", min(sig12), max(sig12), np.mean(sig12)],
            ["sign1", min(sign2), max(sign2), np.mean(sign2)],
        ]
        Pr.prin(tabulate(table))

    def _updateApproximation(self):
        """
        Calculates the Laplace approximation for the posterior.
        It can be defined by two variables: f mode and W at f mode.
        """
        if self._updateApproximationCount == 0:
            return

        if self._is_kernel_zero():
            self._updateApproximationCount = 0
            return

        self._updateApproximationBegin()

        gradEpsStop = 1e-10
        objEpsStop = 1e-8
        gradEpsErr = 1e-3

        self._mean = self._calculateMean()
        m = self._mean

        if self._lasta is None or self._lasta.shape[0] != self._N:
            aprev = np.zeros(self._N)
        else:
            aprev = self._lasta

        fprev = self._rdotK(aprev) + m
        objprev = self._likelihood.log(fprev, self._y) - (fprev - m).dot(aprev) / 2.0
        ii = 0
        line_search = False
        maxIter = 1000
        failed = False
        failedMsg = ""
        while ii < maxIter:
            grad = self._calculateUAGrad(fprev, aprev)
            if np.mean(abs(grad)) < gradEpsStop:
                a = aprev
                f = fprev
                break

            # The following is just a Newton step (eq. (3.18) [1]) to maximize
            # log(p(F|X,y)) over F
            g = self._likelihood.gradient_log(fprev, self._y)
            W = self._calculateW(fprev)
            b = W * (fprev - m) + g

            a = self._calculateUAa(b, W)

            if line_search:
                (f, a, obj) = self._lineSearch(a, aprev, m)
            else:
                f = self._rdotK(a) + m
                obj = self._likelihood.log(f, self._y) - (f - m).dot(a) / 2.0

            if abs(objprev - obj) < objEpsStop:
                grad = self._calculateUAGrad(f, a)
                break
            if obj > objprev:
                fprev = f
                objprev = obj
                aprev = a
            else:
                if line_search:
                    grad = self._calculateUAGrad(fprev, aprev)
                    a = aprev
                    f = fprev
                    break
                line_search = True
            ii += 1

        self._lasta = a

        err = np.mean(abs(grad))
        if err > gradEpsErr:
            failed = True
            failedMsg = "Gradient not too small in the Laplace update approximation.\n"
            failedMsg = (
                failedMsg
                + "Problem in the f mode estimation. |grad|_{mean} = %.6f." % (err,)
            )

        if ii >= maxIter:
            failed = True
            failedMsg = (
                "Laplace update approximation did not converge in less than maxIter."
            )

        if self._debug:
            self._debugUACalls.append(
                self.DebugUACall(
                    ii,
                    grad,
                    not failed,
                    self.beta,
                    self._sig02,
                    self._sig12,
                    self._sign2,
                )
            )

        if failed:
            Pr.prin(
                "Laplace update approximation failed. The failure message is the following."
            )
            Pr.prin(failedMsg)
            sys.exit("Stopping program.")

        self._updateApproximationEnd(f, a)

        self._updateApproximationCount = 0

    def _predict(self, meanstar, kstar, kstarstar, prob):
        self._updateConstants()
        self._updateApproximation()

        if np.isscalar(kstarstar):
            return self._predict_each(meanstar, kstar, kstarstar, prob)

        n = len(kstarstar)
        ps = np.zeros(n)

        for i in range(n):
            ps[i] = self._predict_each(meanstar[i], kstar[i, :], kstarstar[i], prob)

        return ps


class LaplaceGLMM_N1K3(GLMM_N1K3, LaplaceGLMM):
    def __init__(self, link, penalty=None, penalizeBias=False, debug=False):
        GLMM_N1K3.__init__(
            self, penalty=penalty, penalizeBias=penalizeBias, debug=debug
        )
        LaplaceGLMM.__init__(self, link)
        self._link = link

    def _updateApproximationBegin(self):
        self._G01 = self._calculateG01()

    def _calculateUAa(self, b, W):
        A = 1.0 + W * self._sign2
        V = W / A
        Lk = self._calculateLk(self._G01, V)

        Gtb = dot(self._G01.T, b)
        GtV = ddot(self._G01.T, V, left=False)
        LtLGtV = stu(Lk.T, stl(Lk, GtV))
        LtLGtVG = dot(LtLGtV, self._G01)
        bn = self._sign2 * b
        a = (
            b
            + dot(dot(GtV.T, LtLGtVG) - GtV.T, Gtb)
            + dot(GtV.T, dot(LtLGtV, bn))
            - V * bn
        )

        return a

    def _updateApproximationEnd(self, f, a):
        self._f = f
        self._a = a

        self._W = self._calculateW(f)
        self._Wsq = np.sqrt(self._W)

        self._A = 1.0 + self._W * self._sign2
        self._V = self._W / self._A
        self._Lk = self._calculateLk(self._G01, self._V)

    def _updateApproximation(self):
        LaplaceGLMM._updateApproximation(self)

    def _regular_marginal_loglikelihood(self):
        self._updateConstants()
        self._updateApproximation()

        if self._is_kernel_zero():
            return self._likelihood.log(self._mean, self._y)

        (f, a) = (self._f, self._a)

        loglike = self._likelihood.log(f, self._y)

        r = (
            loglike
            - dot(f - self._mean, a) / 2.0
            - sum(np.log(np.diag(self._Lk)))
            - sum(np.log(self._A)) / 2.0
        )
        assert np.isfinite(r), "Not finite regular marginal loglikelihood."

        return r

    def _rmll_gradient(self, optSig02=True, optSig12=True, optSign2=True, optBeta=True):
        self._updateConstants()
        self._updateApproximation()

        (f, a) = (self._f, self._a)
        (W, Wsq) = (self._W, self._Wsq)
        Lk = self._Lk

        m = self._mean
        X = self._X
        G0 = self._G0
        G1 = self._G1
        sign2 = self._sign2
        G01 = self._G01

        # g = self._likelihood.gradient_log(f)
        # a==g

        h = self._likelihood.third_derivative_log(f)

        V = W / self._A

        d = self._dKn()
        G01tV = ddot(G01.T, V, left=False)
        H = stl(Lk, G01tV)
        dkH = self._ldotK(H)
        diags = (
            d
            - sign2**2 * V
            - dotd(G01, dot(dot(G01tV, G01), G01.T))
            - 2.0 * sign2 * dotd(G01, G01tV)
            + dotd(dkH.T, dkH)
        ) * h

        ret = []

        if optSig02:
            dK0a = dot(G0, dot(G0.T, a))
            t = V * dK0a - dot(H.T, dot(H, dK0a))
            dF0 = dK0a - self._rdotK(t)

            LkG01VG0 = dot(H, G0)
            VG0 = ddot(V, G0, left=True)

            ret0 = (
                dot(a, dF0)
                - 0.5 * dot(a, dK0a)
                + dot(f - m, t)
                + 0.5 * np.sum(diags * dF0)
                + -0.5 * trace2(VG0, G0.T)
                + 0.5 * trace2(LkG01VG0.T, LkG01VG0)
            )

            ret.append(ret0)

        if optSig12:
            dK1a = dot(G1, dot(G1.T, a))
            t = V * dK1a - dot(H.T, dot(H, dK1a))
            dF1 = dK1a - self._rdotK(t)

            LkG01VG1 = dot(H, G1)
            VG1 = ddot(V, G1, left=True)

            ret1 = (
                dot(a, dF1)
                - 0.5 * dot(a, dK1a)
                + dot(f - m, t)
                + 0.5 * np.sum(diags * dF1)
                + -0.5 * trace2(VG1, G1.T)
                + 0.5 * trace2(LkG01VG1.T, LkG01VG1)
            )

            ret.append(ret1)

        if optSign2:
            t = V * a - dot(H.T, dot(H, a))
            dFn = a - self._rdotK(t)

            retn = (
                dot(a, dFn)
                - 0.5 * dot(a, a)
                + dot(f - m, t)
                + 0.5 * np.sum(diags * dFn)
                + -0.5 * np.sum(V)
                + 0.5 * trace2(H.T, H)
            )

            ret.append(retn)

        if optBeta:
            t = ddot(V, X, left=True) - dot(H.T, dot(H, X))
            dFbeta = X - self._rdotK(t)

            retbeta = dot(a, dFbeta) + dot(f - m, t)
            for i in range(dFbeta.shape[1]):
                retbeta[i] += 0.5 * np.sum(diags * dFbeta[:, i])

            ret.extend(retbeta)

        ret = np.array(ret)
        assert np.all(
            np.isfinite(ret)
        ), "Not finite regular marginal loglikelihood gradient."

        return ret

    def _predict(self, meanstar, kstar, kstarstar, prob):
        return LaplaceGLMM._predict(self, meanstar, kstar, kstarstar, prob)

    def _predict_each(self, meanstar, kstar, xstarstar, prob):
        """
        Calculates the probability of being 1, or the most probable label
        if prob=False.
        --------------------------------------------------------------------------
        Input:
        meanstar        : input mean.
        kstar           : covariance between provided and prior latent variables.
        xstarstar       : variance of the latent variable.
        prob            : True for probability calculation or False for returning
                          the most probable label.
        """
        a = self._a

        fstarmean = meanstar + kstar.dot(a)

        if prob is False:
            if fstarmean > 0.0:
                return +1.0
            return -1.0

        r0Tr0 = kstar.dot(self._V * kstar)

        r1 = stl(self._Lk, dot(self._G01.T, self._V * kstar))

        fstarvar = xstarstar - r0Tr0 + r1.dot(r1)

        return self._likelihood.intOverGauss(fstarmean, fstarvar)


class LaplaceGLMM_N3K1(GLMM_N3K1, LaplaceGLMM):
    def __init__(self, link, penalty=None, penalizeBias=False, debug=False):
        GLMM_N3K1.__init__(
            self, penalty=penalty, penalizeBias=penalizeBias, debug=debug
        )
        LaplaceGLMM.__init__(self, link)
        self._link = link

    def _updateApproximationBegin(self):
        self._K = np.eye(self._N) * self._sign2
        if self._isK0Set:
            self._K += self._sig02 * (
                self._K0 if self._K0 is not None else dot(self._G0, self._G0.T)
            )
        if self._isK1Set:
            self._K += self._sig12 * (
                self._K1 if self._K1 is not None else dot(self._G1, self._G1.T)
            )

    def _calculateUAa(self, b, W):
        Wsq = np.sqrt(W)
        Ln = self._calculateLn(self._K, Wsq)
        a = b - Wsq * stu(Ln.T, stl(Ln, Wsq * dot(self._K, b)))
        return a

    def _updateApproximationEnd(self, f, a):
        self._f = f
        self._a = a

        self._W = self._calculateW(f)
        self._Wsq = np.sqrt(self._W)

        self._A = 1.0 + self._W * self._sign2
        V = self._W / self._A
        self._Ln = self._calculateLn(self._K, self._Wsq)

    def _regular_marginal_loglikelihood(self):
        self._updateConstants()
        self._updateApproximation()

        if self._is_kernel_zero():
            return self._likelihood.log(self._mean, self._y)

        (f, a) = (self._f, self._a)

        loglike = self._likelihood.log(f, self._y)

        r = loglike - dot(f - self._mean, a) / 2.0 - sum(np.log(np.diag(self._Ln)))
        assert np.isfinite(r), "Not finite regular marginal loglikelihood."

        return r

    def _rmll_gradient(self, optSig02=True, optSig12=True, optSign2=True, optBeta=True):
        self._updateConstants()
        self._updateApproximation()

        W = self._W
        Wsq = self._Wsq
        f = self._f
        K = self._K
        K0 = self._K0
        K1 = self._K1
        m = self._mean
        a = self._a
        Ln = self._Ln
        X = self._X

        LnWsq = stl(Ln, np.diag(Wsq))
        LnWsqK = dot(LnWsq, K)

        d = self._dKn()
        h = self._likelihood.third_derivative_log(f)
        diags = (d - dotd(LnWsqK.T, LnWsqK)) * h

        ret = []

        if optSig02:
            dK0a = dot(K0, a)
            dF0 = dK0a - dot(LnWsqK.T, dot(LnWsq, dK0a))

            r = (
                dot(a, dF0)
                - dot(a, dF0)
                + 0.5 * dot(a, dK0a)
                + 0.5 * np.sum(diags * dF0)
                - 0.5 * trace2(LnWsq.T, dot(LnWsq, K0))
            )

            ret.append(r)

        if optSig12:
            dK1a = dot(K1, a)
            dF1 = dK1a - dot(LnWsqK.T, dot(LnWsq, dK1a))

            r = (
                dot(a, dF1)
                - dot(a, dF1)
                + 0.5 * dot(a, dK1a)
                + 0.5 * np.sum(diags * dF1)
                - 0.5 * trace2(LnWsq.T, dot(LnWsq, K1))
            )

            ret.append(r)

        if optSign2:
            dFn = a - dot(LnWsqK.T, dot(LnWsq, a))

            r = (
                dot(a, dFn)
                - dot(a, dFn)
                + 0.5 * dot(a, a)
                + 0.5 * np.sum(diags * dFn)
                - 0.5 * trace2(LnWsq.T, LnWsq)
            )

            ret.append(r)

        if optBeta:
            dFmb = -dot(LnWsqK.T, dot(LnWsq, X))
            dFb = dFmb + X

            r = dot(a, dFb) - dot(a, dFmb) + 0.5 * np.sum(diags * dFb.T, 1)

            ret += list(r)

        ret = np.array(ret)
        assert np.all(
            np.isfinite(ret)
        ), "Not finite regular marginal loglikelihood gradient."

        return ret

    def _updateApproximation(self):
        LaplaceGLMM._updateApproximation(self)

    def _predict(self, meanstar, kstar, kstarstar, prob):
        return LaplaceGLMM._predict(self, meanstar, kstar, kstarstar, prob)

    def _predict_each(self, meanstar, kstar, xstarstar, prob):
        """
        Calculates the probability of being 1, or the most probable label
        if prob=False.
        --------------------------------------------------------------------------
        Input:
        meanstar        : input mean.
        kstar           : covariance between provided and prior latent variables.
        xstarstar       : variance of the latent variable.
        prob            : True for probability calculation or False for returning
                          the most probable label.
        """

        fstarmean = meanstar + kstar.dot(self._a)

        if prob is False:
            if fstarmean > 0.0:
                return +1.0
            return -1.0

        r = stl(self._Ln, self._Wsq * kstar)

        fstarvar = xstarstar - dot(r, r)

        return self._likelihood.intOverGauss(fstarmean, fstarvar)



# import matplotlib
# matplotlib.use("TKAgg",warn=False)
# import pylab
import pandas as pd
import sys

import numpy as np
import logging
import unittest
import doctest
import os.path
from pysnptools.snpreader import Bed, Pheno, SnpData
from fastlmm.association import single_snp_select
from fastlmm.feature_selection.test import TestFeatureSelection
import platform
import multiprocessing
from pysnptools.util.mapreduce1.runner import LocalMultiProc


class TestSingleSnpSelect(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        from pysnptools.util import create_directory_if_necessary

        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.pythonpath = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "..")
        )
        self.bedbase = os.path.join(
            self.pythonpath, "tests/datasets/all_chr.maf0.001.N300"
        )
        self.phen_fn = os.path.join(
            self.pythonpath, "tests/datasets/phenSynthFrom22.23.N300.randcidorder.txt"
        )
        self.cov_fn = os.path.join(
            self.pythonpath, "tests/datasets/all_chr.maf0.001.covariates.N300.txt"
        )

    tempout_dir = "tempout/single_snp_select"

    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name + ".txt")
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    # Break these tests up into six parts so they can run faster when on cluster.
    def test_sel_plus_pc_h2Search(self):  #!!! rather a big test case
        logging.info("TestSingleSnpSelect sel_plus_pc_h2Search")
        self._sel_plus_pc(None, None, None, count_A1=False)

    def test_sel_plus_pc_h2Search_low(self):  #!!! rather a big test case
        logging.info("TestSingleSnpSelect sel_plus_pc_h2Search_low")
        self._sel_plus_pc(None, True, False, count_A1=False)

    def test_sel_plus_pc_h2Search_full(self):  #!!! rather a big test case
        logging.info("TestSingleSnpSelect sel_plus_pc_h2Search_full")
        self._sel_plus_pc(None, False, True, count_A1=False)

    def test_sel_plus_pc_h2IsHalf(self):  #!!! rather a big test case
        logging.info("TestSingleSnpSelect sel_plus_pc_h2IsHalf")
        self._sel_plus_pc(0.5, None, None, count_A1=False)

    def test_sel_plus_pc_h2IsHalf_low(self):  #!!! rather a big test case
        logging.info("TestSingleSnpSelect sel_plus_pc_h2IsHalf_low")
        self._sel_plus_pc(0.5, True, False, count_A1=False)

    def test_sel_plus_pc_h2IsHalf_full(self):  #!!! rather a big test case
        logging.info("TestSingleSnpSelect sel_plus_pc_h2IsHalf_full")
        self._sel_plus_pc(0.5, False, True, count_A1=False)

    def _sel_plus_pc(self, h2, force_low_rank, force_full_rank, count_A1=None):
        do_plot = False
        use_cache = False

        # define file names
        bed_fn = self.pythonpath + "/tests/datasets/synth/all.bed"
        phen_fn = self.pythonpath + "/tests/datasets/synth/pheno_10_causals.txt"

        pcs_fn = os.path.join(self.tempout_dir, "sel_plus_pc.pcs.txt")
        if not (use_cache and os.path.exists(pcs_fn)):
            from fastlmm.util import compute_auto_pcs

            covar = compute_auto_pcs(bed_fn, count_A1=count_A1)
            logging.info("selected number of PCs: {0}".format(covar["vals"].shape[1]))
            Pheno.write(
                pcs_fn,
                SnpData(iid=covar["iid"], sid=covar["header"], val=covar["vals"]),
            )
        else:
            logging.info("Using top pcs's cache")
            covar = Pheno(pcs_fn)

        runner = LocalMultiProc(multiprocessing.cpu_count(), mkl_num_threads=2)

        logging.info(
            "Working on h2={0},force_low_rank={1},force_full_rank={2}".format(
                h2, force_low_rank, force_full_rank
            )
        )
        result_file_name = "sel_plus_pc_{0}".format(
            "h2IsHalf" if h2 == 0.5 else "h2Search"
        )
        output_file_name = os.path.join(self.tempout_dir, result_file_name) + ".txt"
        results = single_snp_select(
            test_snps=bed_fn,
            G=bed_fn,
            pheno=phen_fn,
            k_list=[
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                20,
                30,
                40,
                50,
                60,
                70,
                80,
                90,
                100,
                125,
                160,
                200,
                250,
                320,
                400,
                500,
                630,
                800,
                1000,
            ],
            h2=h2,
            n_folds=self.pythonpath + "/tests/datasets/synth/DebugEmitFolds.txt",
            covar=covar,
            output_file_name=output_file_name,
            force_low_rank=force_low_rank,
            force_full_rank=force_full_rank,
            GB_goal=2,
            count_A1=False,
            # runner = runner
        )
        logging.info(results.head())
        self.compare_files(results, result_file_name)

    def test_old_sel_plus_pc(self):  #!!! rather a big test case
        if platform.system() == "Darwin":  # Don't run old C code on Mac
            return

        logging.info("TestSingleSnpSelect old_sel_plus_pc")

        from pysnptools.snpreader import Bed
        from fastlmm.util import compute_auto_pcs

        # define file names
        bed_fn = self.pythonpath + "/tests/datasets/synth/all"
        cov_fn = "pcs_cov.txt"

        # run PCgeno
        # TODO: rename to auto_pcs
        result = compute_auto_pcs(bed_fn, output_file_name=cov_fn, count_A1=False)
        logging.info("selected number of PCs: {0}".format(result["vals"].shape[1]))

        # import algorithms
        from fastlmm.util.run_fastlmmc import runFASTLMM, runLMMSELECT

        # set some file paths for fastlmmc
        phen_fn = self.pythonpath + "/tests/datasets/synth/pheno_10_causals.txt"
        out_dir = self.tempout_dir
        fastlmm_path = self.pythonpath + "/external/fastlmmc"

        # consists of two fastlmmc calls, one that does feature selection and one that runs GWAS
        for suffix, logdelta in [("h2IsHalf", 0), ("h2Search", None)]:
            result_file_name = "sel_plus_pc_old_{0}".format(
                "h2IsHalf" if logdelta == 0 else "h2Search"
            )
            runLMMSELECT(
                bed_fn,
                phen_fn,
                out_dir,
                result_file_name,
                bfileSim=bed_fn,
                covar=cov_fn,
                fastlmm_path=fastlmm_path,
                autoSelectCriterionMSE=False,
                excludeByGeneticDistance=1000,
                optLogdelta=logdelta,
            )
            # compare sel_plus_pc_old_h2*.LMMSELECT.out.txt
            short = result_file_name + ".LMMSELECT.out"
            results = pd.read_csv(
                self.tempout_dir + "/" + short + ".txt",
                delimiter="\s",
                comment=None,
                engine="python",
            )
            results["PValue"] = (
                results.Pvalue
            )  # add a new column with different capitalization
            self.compare_files(results, short)

    def compare_files(self, frame, ref_base):
        reffile = TestFeatureSelection.reference_file(
            "single_snp_select/" + ref_base + ".txt"
        )

        # sid_list,pvalue_list = frame['SNP'].values,frame['Pvalue'].values

        # sid_to_pvalue = {}
        # for index, sid in enumerate(sid_list):
        #    sid_to_pvalue[sid] = pvalue_list[index]

        reference = pd.read_csv(reffile, delimiter="\s", comment=None, engine="python")
        if "Pvalue" in reference.columns:
            reference["PValue"] = (
                reference.Pvalue
            )  # add a new column with different capitalization if it is there

        assert len(frame) == len(
            reference
        ), "# of pairs differs from file '{0}'".format(reffile)
        for _, row in reference.iterrows():
            sid = row.SNP
            pvalue = frame[frame["SNP"] == sid].iloc[0].PValue
            assert (
                abs(row.PValue - pvalue) < 1e-5
            ), "pair {0} differs too much from file '{1}'".format(sid, reffile)

    def test_doctest(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
        result = doctest.testmod(sys.modules["fastlmm.association.single_snp_select"])
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():

    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestSingleSnpSelect)
    return unittest.TestSuite([suite1])


if __name__ == "__main__":

    # this import is needed for the runner
    from fastlmm.association.tests.test_single_snp_select import TestSingleSnpSelect

    suites = unittest.TestSuite([getTestSuite()])

    if True:  # Standard test run
        r = unittest.TextTestRunner(failfast=False)
        ret = r.run(suites)
        assert ret.wasSuccessful()
    else:  # Cluster test run

        from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc

        logging.basicConfig(level=logging.INFO)

        from pysnptools.util.mapreduce1.distributabletest import DistributableTest

        # runner = HPC(10, 'RR1-N13-09-H44',r'\\msr-arrays\Scratch\msr-pool\Scratch_Storage4\Redmond',
        #                remote_python_parent=r"\\msr-arrays\Scratch\msr-pool\Scratch_Storage4\REDMOND\carlk\Source\carlk\july_7_14\tests\runs\2014-07-24_15_02_02_554725991686\pythonpath",
        #                update_remote_python_parent=True,
        #                priority="AboveNormal",mkl_num_threads=1)
        runner = Local()
        # runner = LocalMultiProc(taskcount=20,mkl_num_threads=5)
        # runner = LocalInParts(1,2,mkl_num_threads=1) # For debugging the cluster runs
        # runner = Hadoop(100, mapmemory=8*1024, reducememory=8*1024, mkl_num_threads=1, queue="default")
        distributable_test = DistributableTest(suites, "temp_test")
        print(runner.run(distributable_test))

    logging.info("done with testing")
